from typing import Dict
from typing import List
from typing import Optional

from aiohttp import ClientSession
from aiohttp import UnixConnector
from aiohttp_xmlrpc.client import ServerProxy
from aiohttp_xmlrpc.exceptions import ServerError

from supervisor_gateway.config import conf
from supervisor_gateway.supervisor import Faults


class RPC:
    def __init__(self, url: str):
        self.url = url
        self.client: Optional[ServerProxy] = None

    def init_client(self):
        if self.url.startswith("unix://"):
            conn = UnixConnector(path=self.url[7:])
            session = ClientSession(connector=conn)
            client = ServerProxy("http://127.0.0.1/RPC2", client=session)
        else:
            client = ServerProxy(self.url)
        self.client = client

    async def close(self):
        if self.client:
            await self.client.close()
            self.client = None

    async def list_methods(self) -> List[str]:
        return await self.client.system.listMethods()

    async def get_state(self) -> Dict:
        state = await self.client.supervisor.getState()
        return state

    async def reread(self) -> List[List[str]]:
        data = await self.client.supervisor.reloadConfig()
        added, changed, dropped = data[0]
        return [added, changed, dropped]

    async def get_all_process_info(self) -> List[Dict]:
        return await self.client.supervisor.getAllProcessInfo()

    async def get_process_info(self, name: str) -> bool:
        if ":" not in name:
            name = f"{name}:{name}"
        return await self.client.supervisor.getProcessInfo(name)

    async def start_process(self, name: str) -> bool:
        if ":" not in name:
            name = f"{name}:{name}"
        try:
            return await self.client.supervisor.startProcess(name)
        except ServerError as e:
            if e.code == Faults.ALREADY_STARTED:
                return True
            raise

    async def restart_process(self, name: str) -> bool:
        await self.stop_process(name)
        return await self.start_process(name)

    async def stop_process(self, name: str) -> bool:
        if ":" not in name:
            name = f"{name}:{name}"
        try:
            return await self.client.supervisor.stopProcess(name)
        except ServerError as e:
            if e.code == Faults.NOT_RUNNING:
                return True
            raise

    async def stop_process_group(self, name: str) -> bool:
        return await self.client.supervisor.stopProcessGroup(name)

    async def add_process_group(self, name: str) -> bool:
        try:
            return await self.client.supervisor.addProcessGroup(name)
        except ServerError as e:
            if e.code == Faults.ALREADY_ADDED:
                return True
            raise

    async def remove_process_group(self, name: str) -> bool:
        await self.stop_process_group(name)
        return await self.client.supervisor.removeProcessGroup(name)


rpc = RPC(conf.rpc_url)



from typing import List

from pydantic import BaseModel


class SupervisorState(BaseModel):
    state: str


class ProcessInfo(BaseModel):
    name: str
    state: str
    update_time: int = 0


class LocalProcesses(BaseModel):
    total: int
    processes: List[ProcessInfo]



from typing import List

from fastapi import APIRouter

from supervisor_gateway.schema import ProcessInfo
from supervisor_gateway.schema import SupervisorState
from supervisor_gateway.supervisor import format_process_info
from supervisor_gateway.xml_rpc import rpc

router = APIRouter()


@router.get("/state", response_model=SupervisorState)
async def get_rpc_state():
    data = await rpc.get_state()
    return {"state": data["statename"]}


@router.get("/processes", response_model=List[ProcessInfo])
async def list_rpc_processes():
    data = await rpc.get_all_process_info()
    processes = [format_process_info(item) for item in data]
    return processes


@router.get("/processes/{name}", response_model=ProcessInfo)
async def get_rpc_process(name: str):
    item = await rpc.get_process_info(name)
    return format_process_info(item)


@router.post("/processes/{name}/stop")
async def stop_rpc_process(name: str):
    await rpc.stop_process(name)


@router.post("/processes/{name}/start")
async def start_rpc_process(name: str):
    await rpc.start_process(name)


@router.post("/processes/{name}/restart")
async def restart_rpc_process(name: str):
    await rpc.restart_process(name)


@router.post("/processes/{name}/add")
async def add_rpc_process(name: str):
    await rpc.add_process_group(name)


@router.post("/processes/{name}/remove")
async def remove_rpc_process(name: str):
    await rpc.remove_process_group(name)



from fastapi import APIRouter

from supervisor_gateway.local_state import state
from supervisor_gateway.schema import LocalProcesses

router = APIRouter()


@router.get("/processes", response_model=LocalProcesses)
async def list_processes(
    keywords: str = "",
    status: str = "",
    page: int = 1,
    limit: int = 50,
):
    processes = sorted(state.processes.items())
    for keyword in keywords.split(","):
        processes = [(name, process) for name, process in processes if keyword in name]
    if status:
        processes = [
            (name, process) for name, process in processes if status == process["state"]
        ]
    total = len(processes)
    start = max((page - 1) * limit, 0)
    end = start + limit
    return {"total": total, "processes": [p for _, p in processes[start:end]]}



READY = "READY\n"
ACKNOWLEDGED = "RESULT 2\nOK"


class ProcessStates:
    STOPPED = 0
    STARTING = 10
    RUNNING = 20
    BACKOFF = 30
    STOPPING = 40
    EXITED = 100
    FATAL = 200
    UNKNOWN = 1000


class Faults:
    UNKNOWN_METHOD = 1
    INCORRECT_PARAMETERS = 2
    BAD_ARGUMENTS = 3
    SIGNATURE_UNSUPPORTED = 4
    SHUTDOWN_STATE = 6
    BAD_NAME = 10
    BAD_SIGNAL = 11
    NO_FILE = 20
    NOT_EXECUTABLE = 21
    FAILED = 30
    ABNORMAL_TERMINATION = 40
    SPAWN_ERROR = 50
    ALREADY_STARTED = 60
    NOT_RUNNING = 70
    SUCCESS = 80
    ALREADY_ADDED = 90
    STILL_RUNNING = 91
    CANT_REREAD = 92


def get_state_time(process_info: dict) -> int:
    if process_info["state"] in (
        ProcessStates.RUNNING,
        ProcessStates.STARTING,
    ):
        return process_info["start"]
    return process_info["stop"]


def format_process_info(process_info: dict) -> dict:
    return {
        "name": process_info["name"],
        "state": process_info["statename"],
        "update_time": get_state_time(process_info),
    }



from typing import Dict
from typing import List
from typing import Union


class BaseError(Exception):
    code: int = 50000
    name: str = "UNKNOWN_ERROR"
    detail: Union[str, List[Dict]] = ""

    def __init__(self, detail: Union[str, List[Dict]] = ""):
        self.detail = detail

    @property
    def http_code(self) -> int:
        return self.code // 100

    def dict(self) -> dict:
        return {
            "code": self.code,
            "name": self.name,
            "detail": self.detail,
        }


class BadRequestError(BaseError):
    code: int = 40000
    name: str = "BAD_REQUEST"


class InvalidParameterError(BaseError):
    code: int = 40001
    name: str = "INVALID_PARAMETER"


class NotFoundError(BaseError):
    code: int = 40400
    name: str = "PATH_NOT_FOUND"


class ProcessNotFoundError(BaseError):
    code: int = 40401
    name: str = "PROCESS_NOT_FOUND"


class MethodNotAllowedError(BaseError):
    code: int = 40500
    name: str = "METHOD_NOT_ALLOWED"


class InternalError(BaseError):
    name: str = "INTERNAL_ERROR"


class SupervisordError(BaseError):
    code: int = 50001
    name: str = "SUPERVISORD_ERROR"



import asyncio
import sys
from asyncio import StreamReader
from asyncio import StreamReaderProtocol
from asyncio import StreamWriter
from typing import Callable
from typing import Dict
from typing import Union

from supervisor_gateway.log import logger
from supervisor_gateway.supervisor import ACKNOWLEDGED
from supervisor_gateway.supervisor import READY


class StandardStreamReaderProtocol(StreamReaderProtocol):
    def connection_made(self, transport):
        if self._stream_reader._transport is not None:  # noqa
            return
        super().connection_made(transport)


async def write(stream: StreamWriter, msg: str):
    stream.write(msg.encode())
    await stream.drain()


async def read(stream: StreamReader) -> Dict[str, Union[str, Dict]]:
    data = await stream.readline()
    header_line = data.decode()
    event: Dict = {}
    for line in header_line.split():
        k, v = line.split(":", 1)
        event[k] = v
    data = await stream.read(int(event["len"]))
    payload_str = data.decode()
    payload = {}
    for line in payload_str.split():
        k, v = line.split(":", 1)
        payload[k] = v
    event["payload"] = payload
    return event


async def open_connection(
    in_pipe=sys.stdin,
    out_pipe=sys.stdout,
) -> (StreamReader, StreamWriter):
    loop = asyncio.get_event_loop()
    reader = asyncio.StreamReader(loop=loop)
    protocol = StandardStreamReaderProtocol(reader)
    await loop.connect_read_pipe(lambda: protocol, in_pipe)

    out_transport_connect = loop.connect_write_pipe(lambda: protocol, out_pipe)
    out_transport, _ = await out_transport_connect
    writer = asyncio.StreamWriter(out_transport, protocol, reader, loop)
    return reader, writer


class Listener:
    def __init__(self, handler: Callable[[dict], None] = None):
        self.handler = handler
        self.running = False

    def set_handler(self, handler: Callable[[dict], None]):
        self.handler = handler

    def stop(self):
        self.running = False

    async def start(self):
        self.running = True
        logger.info("event_listener started")
        reader, writer = await open_connection()
        while self.running:
            await write(writer, READY)
            event = await read(reader)
            self.handler(event)
            logger.debug(f"event: {event}")
            await write(writer, ACKNOWLEDGED)


listener = Listener()



from typing import Any
from typing import Dict
from typing import List

from supervisor_gateway.log import logger
from supervisor_gateway.supervisor import format_process_info
from supervisor_gateway.time import now


class LocalSate:
    def __init__(self):
        self.processes: Dict[str, Dict] = {}
        self.supervisor: Dict[str, str] = {}

    def update_supervisor(self, running_state: str):
        self.supervisor["state"] = running_state

    def add_process(self, process: Dict):
        self.processes[process["name"]] = format_process_info(process)

    def add_processes(self, processes: List[Dict]):
        for p in processes:
            self.add_process(p)

    def clean(self):
        self.processes = {}
        self.supervisor = {}

    def event_handler(self, event: Dict[str, Any]):
        if "eventname" not in event:
            return
        if event["eventname"].startswith("SUPERVISOR_STATE_CHANGE_"):
            state_name = event["eventname"][24:]
            self.update_supervisor(state_name)
        elif event["eventname"].startswith("PROCESS_STATE_"):
            state_name = event["eventname"][14:]
            name = event["payload"]["processname"]
            info = {
                "name": name,
                "state": state_name,
                "update_time": now(),
            }
            self.processes[name] = info
        elif event["eventname"] == "PROCESS_GROUP_REMOVED":
            name = event["payload"]["groupname"]
            try:
                del self.processes[name]
            except KeyError:
                logger.error(f"process {name} not found, event: {event}")
        else:
            logger.warning(f"unknown event: {event}")


state = LocalSate()



import logging
from logging.handlers import RotatingFileHandler

from supervisor_gateway.config import conf

hdl = RotatingFileHandler(
    filename=conf.log_file,
    maxBytes=50 * 1024 * 1024,
    backupCount=3,
)
hdl.setFormatter(
    logging.Formatter(
        "[%(asctime)s] (%(levelname)s):%(filename)s:%(funcName)s:%(lineno)d: %(message)s"
    )
)
logger = logging.getLogger("sg")
logger.setLevel(conf.log_level)
logger.addHandler(hdl)



from asyncio import get_event_loop

from aiohttp_xmlrpc.exceptions import ServerError as RPCServerError
from fastapi import FastAPI
from fastapi.exceptions import RequestValidationError
from fastapi.requests import Request
from fastapi.responses import JSONResponse
from starlette.exceptions import HTTPException
from starlette.status import HTTP_400_BAD_REQUEST
from starlette.status import HTTP_404_NOT_FOUND
from starlette.status import HTTP_405_METHOD_NOT_ALLOWED

from supervisor_gateway.api.local import router as local_router
from supervisor_gateway.api.rpc import router as rpc_router
from supervisor_gateway.error import BadRequestError
from supervisor_gateway.error import InternalError
from supervisor_gateway.error import InvalidParameterError
from supervisor_gateway.error import MethodNotAllowedError
from supervisor_gateway.error import NotFoundError
from supervisor_gateway.error import ProcessNotFoundError
from supervisor_gateway.log import logger
from supervisor_gateway.supervisor import Faults


def get_app() -> FastAPI:
    app_ = FastAPI()
    app_.include_router(local_router, tags=["local"])
    app_.include_router(rpc_router, prefix="/rpc", tags=["rpc"])
    return app_


app = get_app()


@app.on_event("startup")
async def on_startup():
    from supervisor_gateway.event_listener import listener
    from supervisor_gateway.local_state import state
    from supervisor_gateway.xml_rpc import rpc

    rpc.init_client()

    supervisor_state = await rpc.get_state()
    state.update_supervisor(supervisor_state["statename"])
    process_list = await rpc.get_all_process_info()
    state.add_processes(process_list)

    listener.set_handler(state.event_handler)
    loop = get_event_loop()
    loop.create_task(listener.start())
    logger.debug("on_startup finished")


@app.on_event("shutdown")
async def on_shutdown():
    from supervisor_gateway.event_listener import listener
    from supervisor_gateway.xml_rpc import rpc

    listener.stop()
    await rpc.close()
    logger.debug("on_shutdown finished")


@app.exception_handler(RPCServerError)
async def xml_rpc_exception_handler(request: Request, exc: RPCServerError):
    if exc.code == Faults.BAD_NAME:
        err = ProcessNotFoundError()
    else:
        err = InternalError()
        logger.exception(f"{request.method} {request.url} got XMLServerError.")
    return JSONResponse(
        status_code=err.http_code,
        content=err.dict(),
    )


@app.exception_handler(RequestValidationError)
async def request_validation_exception_handler(
    request: Request, exc: RequestValidationError
):
    err = InvalidParameterError(exc.errors())
    return JSONResponse(
        status_code=err.http_code,
        content=err.dict(),
    )


@app.exception_handler(HTTPException)
async def http_exception_handler(request: Request, exc: HTTPException):
    if exc.status_code == HTTP_400_BAD_REQUEST:
        err = BadRequestError(exc.detail)
    elif exc.status_code == HTTP_404_NOT_FOUND:
        err = NotFoundError(exc.detail)
    elif exc.status_code == HTTP_405_METHOD_NOT_ALLOWED:
        err = MethodNotAllowedError(exc.detail)
    else:
        logger.exception(f"{request.method} {request.url} got unknown error.")
        err = InternalError()
    return JSONResponse(
        status_code=err.http_code,
        content=err.dict(),
    )



from setuptools import setup, find_packages

from supervisor_gateway import __version__

extras_require = {}
install_requires = []
with open("requirements.txt") as f:
    lines = f.readlines()
    lines = [line.strip() for line in lines if line.strip()]
    i = 0
    extra_type = ""
    for line in lines:
        if "#" == line[0]:
            extra_type = line[1:].strip("#").strip()
            extras_require[extra_type] = []
            continue
        if extra_type:
            extras_require[extra_type].append(line)
        else:
            install_requires.append(line)


def get_long_description():
    """
    Return the README.
    """
    with open("README.md", encoding="utf8") as f:
        return f.read()


setup(
    name="supervisor-gateway",
    version=__version__,
    url="https://github.com/ppd0705/supervisor-gateway",
    author="ppd0705",
    author_email="ppd0705@icloud.com",
    description="An RESTful supervisor gateway with paginated and cached process info",
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    packages=find_packages(exclude=["tests*"]),
    extras_require=extras_require,
    install_requires=install_requires,
    classifiers=[
        "Intended Audience :: Developers",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: System :: Boot",
        "Topic :: System :: Monitoring",
        "Topic :: System :: Systems Administration",
    ],
    python_requires=">=3.8",
)



import os


class Conf:
    def __init__(self):
        # listen addr
        self.host: str = os.environ.get("SG_HOST") or "localhost"
        self.port: int = int(os.environ.get("SG_PORT") or "1234")

        # supervisor rpc url
        self.rpc_url: str = os.environ.get("SG_RPC") or "http://localhost:9011/RPC2"

        # log
        self.log_level: str = os.environ.get("SG_LOG_LEVEL") or "DEBUG"
        self.log_file: str = os.environ.get("SG_LOG_FILE") or "supervisor_gateway.log"

        # std file
        self.stdin: str = os.environ.get("SG_STDIN") or "/dev/stdin"
        self.stdout: str = os.environ.get("SG_STDOUT") or "/dev/stdout"


conf = Conf()



import pathlib

import uvicorn
from uvicorn.main import LOGGING_CONFIG

from supervisor_gateway.config import conf
from supervisor_gateway.main import app

if __name__ == "__main__":
    path = pathlib.Path(conf.log_file)
    unicorn_path = path.parent / "unicorn.log"
    access_path = path.parent / "unicorn.access.log"
    LOGGING_CONFIG["handlers"] = {
        "default": {
            "formatter": "default",
            "class": "logging.handlers.RotatingFileHandler",
            "filename": str(unicorn_path),
            "maxBytes": 50 * 1024 * 1024,
            "backupCount": 3,
        },
        "access": {
            "formatter": "access",
            "class": "logging.handlers.RotatingFileHandler",
            "filename": str(access_path),
            "maxBytes": 50 * 1024 * 1024,
            "backupCount": 3,
        },
    }
    uvicorn.run(app, host=conf.host, port=conf.port)
